@@ -7,13 +7,27 @@
 #include "rimestate.h"
 #include "rimecandidate.h"
 #include "rimeengine.h"
+#include <cstdint>
+#include <fcitx-utils/capabilityflags.h>
+#include <fcitx-utils/i18n.h>
+#include <fcitx-utils/key.h>
+#include <fcitx-utils/keysym.h>
 #include <fcitx-utils/stringutils.h>
 #include <fcitx-utils/textformatflags.h>
 #include <fcitx-utils/utf8.h>
 #include <fcitx/candidatelist.h>
+#include <fcitx/event.h>
 #include <fcitx/inputcontext.h>
 #include <fcitx/inputpanel.h>
+#include <fcitx/text.h>
+#include <fcitx/userinterface.h>
+#include <functional>
+#include <iterator>
+#include <memory>
+#include <optional>
 #include <rime_api.h>
+#include <string>
+#include <utility>
 
 namespace fcitx {
 
@@ -21,10 +35,8 @@ namespace {
 
 bool emptyExceptAux(const InputPanel &inputPanel) {
 
-    return inputPanel.preedit().size() == 0 &&
-           inputPanel.preedit().size() == 0 &&
-           (!inputPanel.candidateList() ||
-            inputPanel.candidateList()->size() == 0);
+    return inputPanel.preedit().empty() && inputPanel.preedit().empty() &&
+           (!inputPanel.candidateList() || inputPanel.candidateList()->empty());
 }
 } // namespace
 
@@ -92,8 +104,18 @@ std::string RimeState::subModeLabel() {
     return result;
 }
 
+void RimeState::toggleLatinMode() {
+    auto *api = engine_->api();
+    if (api->is_maintenance_mode()) {
+        return;
+    }
+
+    Bool oldValue = api->get_option(session(), RIME_ASCII_MODE);
+    api->set_option(session(), RIME_ASCII_MODE, !oldValue);
+}
+
 void RimeState::setLatinMode(bool latin) {
-    auto api = engine_->api();
+    auto *api = engine_->api();
     if (api->is_maintenance_mode()) {
         return;
     }
@@ -101,7 +123,7 @@ void RimeState::setLatinMode(bool latin) {
 }
 
 void RimeState::selectSchema(const std::string &schema) {
-    auto api = engine_->api();
+    auto *api = engine_->api();
     if (api->is_maintenance_mode()) {
         return;
     }
@@ -111,7 +133,23 @@ void RimeState::selectSchema(const std::string &schema) {
 }
 
 void RimeState::keyEvent(KeyEvent &event) {
-    auto api = engine_->api();
+    auto *ic = event.inputContext();
+    // For key-release, composeResult will always be empty string, which feed
+    // into engine directly.
+    std::string composeResult;
+    if (!event.key().states().testAny(
+            KeyStates{KeyState::Ctrl, KeyState::Super}) &&
+        !event.isRelease()) {
+        auto compose =
+            engine_->instance()->processComposeString(&ic_, event.key().sym());
+        if (!compose) {
+            event.filterAndAccept();
+            return;
+        }
+        composeResult = *compose;
+    }
+
+    auto *api = engine_->api();
     if (api->is_maintenance_mode()) {
         return;
     }
@@ -133,26 +171,44 @@ void RimeState::keyEvent(KeyEvent &event) {
         // IBUS_RELEASE_MASK
         intStates |= (1 << 30);
     }
-    auto result = api->process_key(session, event.rawKey().sym(), intStates);
+    if (!composeResult.empty()) {
+        event.filterAndAccept();
+        auto length = utf8::lengthValidated(composeResult);
+        bool result = false;
+        if (length == 1) {
+            auto c = utf8::getChar(composeResult);
+            auto sym = Key::keySymFromUnicode(c);
+            if (sym != FcitxKey_None) {
+                result = api->process_key(session, sym, intStates);
+            }
+        }
+        if (!result) {
+            commitPreedit(ic);
+            ic->commitString(composeResult);
+            clear();
+        }
+    } else {
+        auto result =
+            api->process_key(session, event.rawKey().sym(), intStates);
+        if (result) {
+            event.filterAndAccept();
+        }
+    }
 
-    auto ic = event.inputContext();
     RIME_STRUCT(RimeCommit, commit);
     if (api->get_commit(session, &commit)) {
         ic->commitString(commit.text);
         api->free_commit(&commit);
+        engine_->instance()->resetCompose(ic);
     }
 
     updateUI(ic, event.isRelease());
-
-    if (result) {
-        event.filterAndAccept();
-    }
 }
 
 #ifndef FCITX_RIME_NO_SELECT_CANDIDATE
 void RimeState::selectCandidate(InputContext *inputContext, int idx,
                                 bool global) {
-    auto api = engine_->api();
+    auto *api = engine_->api();
     if (api->is_maintenance_mode()) {
         return;
     }
@@ -176,7 +232,7 @@ void RimeState::selectCandidate(InputContext *inputContext, int idx,
 
 bool RimeState::getStatus(
     const std::function<void(const RimeStatus &)> &callback) {
-    auto api = engine_->api();
+    auto *api = engine_->api();
     auto session = this->session();
     if (!session) {
         return false;
@@ -190,7 +246,8 @@ bool RimeState::getStatus(
     return true;
 }
 
-Text preeditFromRimeContext(const RimeContext &context, TextFormatFlags flag) {
+Text preeditFromRimeContext(const RimeContext &context, TextFormatFlags flag,
+                            TextFormatFlags highlightFlag) {
     Text preedit;
 
     do {
@@ -218,7 +275,7 @@ Text preeditFromRimeContext(const RimeContext &context, TextFormatFlags flag) {
                 std::string(
                     &context.composition.preedit[context.composition.sel_start],
                     &context.composition.preedit[context.composition.sel_end]),
-                flag | TextFormatFlag::HighLight);
+                flag | highlightFlag);
         }
 
         /* remaining input to convert */
@@ -243,13 +300,14 @@ void RimeState::updatePreedit(InputContext *ic, const RimeContext &context) {
 
     switch (mode) {
     case PreeditMode::No:
-        ic->inputPanel().setPreedit(
-            preeditFromRimeContext(context, TextFormatFlag::NoFlag));
+        ic->inputPanel().setPreedit(preeditFromRimeContext(
+            context, TextFormatFlag::NoFlag, TextFormatFlag::NoFlag));
+        ic->inputPanel().setClientPreedit(Text());
         break;
     case PreeditMode::CommitPreview: {
-        ic->inputPanel().setPreedit(
-            preeditFromRimeContext(context, TextFormatFlag::NoFlag));
-        if (context.commit_text_preview) {
+        ic->inputPanel().setPreedit(preeditFromRimeContext(
+            context, TextFormatFlag::NoFlag, TextFormatFlag::NoFlag));
+        if (context.composition.length > 0 && context.commit_text_preview) {
             Text clientPreedit;
             clientPreedit.append(context.commit_text_preview,
                                  TextFormatFlag::Underline);
@@ -259,11 +317,17 @@ void RimeState::updatePreedit(InputContext *ic, const RimeContext &context) {
                 clientPreedit.setCursor(clientPreedit.textLength());
             }
             ic->inputPanel().setClientPreedit(clientPreedit);
+        } else {
+            ic->inputPanel().setClientPreedit(Text());
         }
     } break;
     case PreeditMode::ComposingText: {
-        Text clientPreedit =
-            preeditFromRimeContext(context, TextFormatFlag::Underline);
+        const TextFormatFlag highlightFlag =
+            *engine_->config().preeditCursorPositionAtBeginning
+                ? TextFormatFlag::HighLight
+                : TextFormatFlag::NoFlag;
+        Text clientPreedit = preeditFromRimeContext(
+            context, TextFormatFlag::Underline, highlightFlag);
         if (*engine_->config().preeditCursorPositionAtBeginning) {
             clientPreedit.setCursor(0);
         }
@@ -280,7 +344,7 @@ void RimeState::updateUI(InputContext *ic, bool keyRelease) {
     bool oldEmptyExceptAux = emptyExceptAux(inputPanel);
 
     do {
-        auto api = engine_->api();
+        auto *api = engine_->api();
         if (api->is_maintenance_mode()) {
             return;
         }
@@ -330,13 +394,13 @@ void RimeState::updateUI(InputContext *ic, bool keyRelease) {
 void RimeState::release() { session_.reset(); }
 
 void RimeState::commitPreedit(InputContext *ic) {
-    if (auto api = engine_->api()) {
+    if (auto *api = engine_->api()) {
         RIME_STRUCT(RimeContext, context);
         auto session = this->session();
         if (!api->get_context(session, &context)) {
             return;
         }
-        if (context.commit_text_preview) {
+        if (context.composition.length > 0 && context.commit_text_preview) {
             ic->commitString(context.commit_text_preview);
         }
         api->free_context(&context);
@@ -356,7 +420,7 @@ void RimeState::snapshot() {
         if (savedCurrentSchema_.empty()) {
             return;
         }
-        auto &optionActions = engine_->optionActions();
+        const auto &optionActions = engine_->optionActions();
         auto iter = optionActions.find(savedCurrentSchema_);
         if (iter == optionActions.end()) {
             return;
